// Copyright (C) The Arvados Authors. All rights reserved.
//
// SPDX-License-Identifier: AGPL-3.0

import { Dispatch, MiddlewareAPI } from "redux";
import { DataExplorerMiddlewareService, dataExplorerToListParams, listResultsToDataExplorerItemsMeta } from "store/data-explorer/data-explorer-middleware-service";
import { RootState } from "store/store";
import { ServiceRepository } from "services/services";
import { snackbarActions, SnackbarKind } from 'store/snackbar/snackbar-actions';
import { DataExplorer, getDataExplorer } from "store/data-explorer/data-explorer-reducer";
import { FilterBuilder } from 'services/api/filter-builder';
import { updateResources } from 'store/resources/resources-actions';
import { getCurrentGroupDetailsPanelUuid, GroupMembersPanelActions } from 'store/group-details-panel/group-details-panel-actions';
import { LinkClass } from 'models/link';
import { ResourceKind } from 'models/resource';
import { progressIndicatorActions } from 'store/progress-indicator/progress-indicator-actions';

export class GroupDetailsPanelMembersMiddlewareService extends DataExplorerMiddlewareService {

    constructor(private services: ServiceRepository, id: string) {
        super(id);
    }

    async requestItems(api: MiddlewareAPI<Dispatch, RootState>) {
        const dataExplorer = getDataExplorer(api.getState().dataExplorer, this.getId());
        const groupUuid = getCurrentGroupDetailsPanelUuid(api.getState().properties);
        if (!dataExplorer || !groupUuid) {
            // Noop if data explorer refresh is triggered from another panel
            return;
        } else {
            try {
                api.dispatch(progressIndicatorActions.START_WORKING(this.getId()));
                const groupResource = await this.services.groupsService.get(groupUuid);
                api.dispatch(updateResources([groupResource]));

                const permissionsIn = await this.services.permissionService.list(getParams(dataExplorer, groupUuid));
                api.dispatch(updateResources(permissionsIn.items));

                api.dispatch(GroupMembersPanelActions.SET_ITEMS({
                    ...listResultsToDataExplorerItemsMeta(permissionsIn),
                    items: permissionsIn.items.map(item => item.uuid),
                }));

                const usersIn = await this.services.userService.list({
                    filters: new FilterBuilder()
                        .addIn('uuid', permissionsIn.items
                            .filter((item) => item.tailKind === ResourceKind.USER)
                            .map(item => item.tailUuid))
                        .getFilters(),
                    count: "none"
                });
                api.dispatch(updateResources(usersIn.items));

                const projectsIn = await this.services.projectService.list({
                    filters: new FilterBuilder()
                        .addIn('uuid', permissionsIn.items
                            .filter((item) => item.tailKind === ResourceKind.PROJECT)
                            .map(item => item.tailUuid))
                        .getFilters(),
                    count: "none"
                });
                api.dispatch(updateResources(projectsIn.items));
            } catch (e) {
                api.dispatch(couldNotFetchGroupDetailsContents());
            } finally {
                api.dispatch(progressIndicatorActions.STOP_WORKING(this.getId()));
            }
        }
    }
}

export const getParams = (dataExplorer: DataExplorer, groupUuid: string) => ({
    ...dataExplorerToListParams(dataExplorer),
    filters: getFilters(groupUuid),
});

export const getFilters = (groupUuid: string) => {
    const filters = new FilterBuilder()
        .addEqual('head_uuid', groupUuid)
        .addEqual('link_class', LinkClass.PERMISSION)
        .getFilters();

    return filters;
};

const couldNotFetchGroupDetailsContents = () =>
    snackbarActions.OPEN_SNACKBAR({
        message: 'Could not fetch group members.',
        kind: SnackbarKind.ERROR
    });
